#include <QApplication>
#include <q3network.h>

#include "flagrowobj.h"
#include "mainwindow.h"
#include "options.h"
#include "settings.h"
#include "version.h"

// Global variables
TextEditor *textEditor;			// used in Constr. of LinkableMapObj
								// initialized in mainwindow
QString vymName(__VYM_NAME);
QString vymVersion(__VYM_VERSION);
QString vymBuildDate(__VYM_BUILD_DATE);

Main *mainWindow;				// used in BranchObj::select()								
QString tmpVymDir;				// All temp files go there, created in mainwindow
QString clipboardDir;			// Clipboard used in all mapEditors
QDir vymBaseDir;				// Containing all styles, scripts, images, ...
QDir lastImageDir;
QDir lastFileDir;
QString iconPath;				// Pointing to icons used for toolbars
QString flagsPath;				// Pointing to flags
bool clipboardEmpty;			
FlagRowObj *systemFlagsDefault;	// used to copy from in LinkableMapObj
FlagRowObj *standardFlagsDefault;


Settings settings ("InSilmaril","vym"); // Organization, Application name

Options options;
ImageIO imageIO;

int statusbarTime=3500;

int main(int argc, char* argv[])
{
	//Q_INIT_RESOURCE (application);

    QApplication app(argc,argv);


	// Reading and initializing options commandline options
	options.add ("version", SwitchOption, "v","version");
	options.add ("local", SwitchOption, "l", "local");
	options.add ("help", SwitchOption, "h", "help");
	options.add ("quit", SwitchOption, "q", "quit");
	options.add ("test", StringOption, "t", "test");
	options.setHelpText (
		"VYM - View Your Mind\n"
		"--------------------\n\n"
		"Information about vym can be found in vym.pdf,\n"
		"which should be part of the vym package.\n"
		"It is also available at the project homepage:\n\n"
		"http://www.InSilmaril.de/vym\n");
	if (options.parse())
	{
		cout << endl << options.getHelpText().ascii()<<endl;
		return 1;
	}

	if (options.isOn ("version"))
	{
		cout << "vym version "<<__VYM_VERSION
			<<"  (c) 2004-2006 Uwe Drechsel - "
			<<__VYM_BUILD_DATE<<endl;
			
		return 0;	
	}		
	
	// Use /usr/share/vym or /usr/local/share/vym or . ?
	// First try options
	if (options.isOn ("local"))
	{
		vymBaseDir.setPath (vymBaseDir.currentDirPath());
	} else
	// then look for environment variable
	if (getenv("VYMHOME")!=0)
	{
		vymBaseDir.setPath (getenv("VYMHOME"));
	} else
	// ok, let's find my way on my own
	{
		#if defined (Q_OS_MACX)
			vymBaseDir.setPath(vymBaseDir.currentDirPath() +"/vym.app/Contents/Resources");

		#else
			vymBaseDir.setPath ("/usr/share/vym");
			if (!vymBaseDir.exists())
			{
				vymBaseDir.setPath ("/usr/local/share/vym");
				if (!vymBaseDir.exists())
					vymBaseDir.setPath(vymBaseDir.currentDirPath() );
			}		
		#endif
	}

	iconPath=vymBaseDir.path()+"/icons/";
	flagsPath=vymBaseDir.path()+"/flags/";

	// Some directories
	lastImageDir=QDir().current();
	lastFileDir=QDir().current();

	if (options.isOn ("help"))
	{
		cout << options.getHelpText().ascii()<<endl;
		return 0;	
	}	

    q3InitNetworkProtocols();


	// Initialize translations
	QTranslator translator (0);
	translator.load( QString("vym_")+QTextCodec::locale(), vymBaseDir.path() + "/lang");
    app.installTranslator( &translator );

	// Initializing the row of system flags
	// is done in first call to MapEditor(),
	// because we need at least one canvas first
	systemFlagsDefault=NULL;
	standardFlagsDefault=NULL;

	// Initialize window of TextEditor
	textEditor = new TextEditor();
	textEditor->setIcon (QPixmap (iconPath+"vym-editor.png"));
	if (textEditor->showWithMain()) textEditor->show();

	// Initialize mainwindow 
    Main m;
	//m.resize(m.sizeHint());
	m.setIcon (QPixmap (iconPath+"vym-48x48.png"));
	m.show();
	m.fileNew();
	m.loadCmdLine();

	// For benchmarking we may want to quit immediatly after drawing
	if (options.isOn ("quit"))
	{
		return 0;
	}	

    QObject::connect( &app, SIGNAL(lastWindowClosed()), &app, SLOT(quit()) );

    return app.exec();
}

